module "rmgr":

/*
 * RMGR Kernel Ressource Manager (c) 1996 C.Schroeter (clausi@chemie.fu-berlin.de)
 * 
 * The rmgr provides a ressource database feature (similar to the Xressource)
 * to the kernel. Each ressource can have its own access permissions if accessed
 * from the user interface. 
 * If a ressource is accessed a callback can be invoked so actions can be invoked
 * depending on a ressource's value. Callbacks cannot be specified 
 * from the user interface.
 *
*/

driver "rmgr" [major=29] {
  
  include Sys.ddl

  /* this includes the other ddl's from ./ddlsrc */
  include types.ddl
  include init.ddl
  include install.ddl
  include lookup.ddl
  include destroy.ddl
  include delete.ddl
  include info.ddl
  include strsave.ddl
  include hash.ddl
  include perm.ddl
  include callback.ddl
  include lock.ddl
  include errors.ddl

  /* the core method */
  core {
  init /{ 
	  RCS_entry *np,v;
          extern void TestCallback();
          printk("Ressource Manager initializing\n");
          RCS_init();

  }/
  cleanup /{
	printk("Ressource Manager unloading\n");
	RCS_destroy();	
  }/
  }
  open /{
  }/
  close /{ 
     /*RCS_info();*/ 
  }/
  /* 
   * This are the definitions for the user interface
   * note that each routine name begins with rmgr_
   *
   */
  ioctl { /*
           * Get the type field
           */
	  func GetType(char *name, int len) /{
		char *buf;
		RCS_entry *np;
                DBGin("GetType");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
		np = RCS_lookup(buf); kfree(buf);
		if( np == NULL ){ 
			Error(-ENOTFOUND);
			DBGout(); return -1;
		}
		DBGout(); return np->type;
	  }/
          /* 
           * rmgr_Chmod
           * Change the access permisions of a ressource
           */
	  func Chmod(char *name, int len, int mode) /{
		  char *buf;
	          char *buf2;
		  RCS_entry entry;
		  RCS_entry *np;
                  DBGin("SetInt");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s mode=0x%x",buf,mode));
		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     RCS_lock(np);
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			Error(-ENOPERM); kfree(buf);
			DBGout(); return -1;
		     }				
		     np->mflags = mode;
		     RCS_unlock(np);
                  }
		  kfree(buf);
		  DBGout(); return np->mflags;
	  }/
          /*
           * rmgr_SetString
           * Set or Create a String Ressource
           *
           */
	  func SetString(char *name, int len,char *sbuf, int slen) /{
		  char *buf;
	          char *buf2;
		  RCS_entry entry;
		  RCS_entry *np;
                  DBGin("SetString");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
		  if( (buf2 = (char *) %<driver.name>_GetFromFS((void *) sbuf,slen)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s string=%s",buf,buf2));
		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     if( np->type != RCS_STRING ){
			DBGprint(DBG_DATA,("type conflict"));
			Error(-ENOTYPE);kfree(buf);kfree(buf2);
			DBGout(); return -1;
		     }
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			Error(-ENOPERM);kfree(buf);kfree(buf2);
			DBGout(); return -1;
		     }
                     RCS_lock(np);
		     entry.type = RCS_STRING;
		     entry.uval.sval = buf2;
                     entry.mflags = np->mflags;
		     RCS_install(buf,&entry);
		     RCS_unlock(np);
                  } else {
		     entry.type = RCS_STRING;
		     entry.uval.sval = buf2;
                     entry.mflags = O_RD | O_WR;
		     RCS_install(buf,&entry);
   		     DBGprint(DBG_DATA,("wrote %s as %s",buf,entry.uval.sval));
                  }
		  kfree(buf);
                  kfree(buf2);
		  DBGout(); return 1;
	  }/
          /*
           * rmgr_GetString
           * Gets a string Ressource
           * 
           */
          func GetString(char *name, int len,char *sbuf) /{
		  char *buf;
		  RCS_entry entry;
		  RCS_entry *np;
                  int slen=0;
                  DBGin("GetString");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("Lookup String name=%s",buf));
		  np = RCS_lookup(buf);
		  if( np == NULL ){
			DBGprint(DBG_DATA,("entry %s not found",buf));
			Error(-ENOTFOUND);kfree(buf);
			DBGout(); return -1;
                  }
                  kfree(buf);
		  RCS_lock(np);
                  if( ! RCS_CheckPerm(np,RCS_PERM_READ)){
			DBGprint(DBG_DATA,("read permission denied"));
			Error(-ENOPERM);
			DBGout(); return -1;
	          }				
		  if(( slen = strlen(np->uval.sval)) > 0){
                        %<driver.name>_PutToFS(sbuf,np->uval.sval,slen);
                  }
		  RCS_unlock(np);
		  DBGout(); return slen;
	  }/
          /*
           * rmgr_SetInt
           * set or create a Integer ressource
           * 
           */
	  func SetInt(char *name, int len, int aint) /{
		  char *buf;
		  RCS_entry entry;
		  RCS_entry *np;
		  extern int umask;

                  DBGin("SetInt");
		  if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                  }
	          DBGprint(DBG_DATA,("name=%s int=%d",buf,aint));

		  if( (np = RCS_lookup(buf)) != NULL ){ 
		     if( np->type != RCS_INT && np->type != RCS_IPTR ){
			DBGprint(DBG_DATA,("type conflict %d",np->type));
			Error(-ENOTYPE);kfree(buf);
			DBGout(); return -1;
		     }
                     if( ! RCS_CheckPerm(np,RCS_PERM_WRITE)){
			DBGprint(DBG_DATA,("write permission denied"));
			DBGout(); return -1;
		     }
		     RCS_lock(np);
                     if( np->type == RCS_INT )
   		           np->uval.ival = aint;
                     else
			   *np->uval.iptr = aint;				
		     RCS_install(buf,np);
		     RCS_unlock(np);
                  } else {
		     entry.type = RCS_INT;
		     entry.uval.ival = aint;
                     entry.mflags = umask;
		     RCS_install(buf,&entry);
  		     DBGprint(DBG_DATA,("wrote %s as %d",buf,entry.uval.ival));
                  }
		  kfree(buf);
		  DBGout();return 1;
	  }/
          /*
           * rmgr_GetInt
           * returns the value of a integer ressource
           * 
           */
	  func GetInt(char *name,int len) /{
		char *buf;
		RCS_entry *np;
                DBGin("GetInt");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
		np = RCS_lookup(buf);
		if( np == NULL ){
			DBGprint(DBG_DATA,("entry not found"));
			Error(-ENOTFOUND);kfree(buf);
			DBGout(); return -1;
                }
                DBGprint(DBG_DATA,("name=%s int=%d",buf,np->uval.ival));
		kfree(buf);
		RCS_lock(np);
                if( ! RCS_CheckPerm(np,RCS_PERM_READ)){
			DBGprint(DBG_DATA,("read permission denied"));
			DBGout(); return -1;
		}				
                RCS_unlock(np);
		DBGout(); 
		if(np->type == RCS_INT)
			return np->uval.ival;
		else
			return *np->uval.iptr;
	  }/
          /*
           * rmgr_Delete
           * deletes a ressource
           * 
           */
	  func Delete(char *name,int len) /{
		char *buf;
                RCS_entry *np;
                DBGin("Delete");
		if( (buf = (char *) %<driver.name>_GetFromFS((void *) name,len)) == NULL ){
                        DBGout();return -EIO;
                }
                if(( np = RCS_lookup(buf)) != NULL ){
                 if( ! RCS_CheckPerm(np,RCS_PERM_WRITE) && !(np->mflags & PERM_STICKY) ){
			DBGprint(DBG_DATA,("delete permission denied"));
			Error(-ENOPERM);
			DBGout(); return -1;
		 }				
	         RCS_delete(buf);
                }
                kfree(buf);
		DBGout(); return 1;		
	  }/
	  func Info() /{
		DBGin("Info");
		RCS_info();
		DBGout();
	  }/
	}

}

